// generated by verigen.pl on Thu Nov 29 16:02:07 2018
// input file: mesi.cfg

module MESI(
	clk, res_n, addr, cp, rd, wr, // inputs
	mesi, index // outputs
 );

input wire clk;
input wire res_n;
input wire [31:0] addr;
input wire cp;
input wire rd;
input wire wr;

output reg [3:0] mesi;
output reg [5:0] index;


  //
 // parameter constants 
//---------------------------
parameter ZERO = 4'b0000;
parameter ZERO_ADDR = 6'b000000;
parameter MODIFIED = 4'b1000;
parameter EXCLUSIVE = 4'b0100;
parameter SHARED = 4'b0010;
parameter INVALID = 4'b0001;


  //
 // state encoding constants 
//---------------------------
parameter [2:0] ID	 = 3'b000;
parameter [2:0] RD	 = 3'b001;
parameter [2:0] WR	 = 3'b010;
parameter [2:0] rRD	 = 3'b101;
parameter [2:0] rWR	 = 3'b110;

reg [2:0] STATE, NEXT_STATE;


  //
 // memory instantiations 
//---------------------------
wire MESI_RAM_clk;
wire MESI_RAM_res;
reg MESI_RAM_wr;
reg [5:0] MESI_RAM_wr_addr;
reg [3:0] MESI_RAM_wr_data;
reg MESI_RAM_rd;
reg [5:0] MESI_RAM_rd_addr;
reg MESI_RAM_status;
wire [3:0] MESI_RAM_rd_data;
assign MESI_RAM_clk = clk;
assign MESI_RAM_res_n = res_n;
RAM #(4,64) MESI_RAM(.clk(MESI_RAM_clk), .res_n(MESI_RAM_res), 
	.wr(MESI_RAM_wr), .wr_addr(MESI_RAM_wr_addr[5:0]), .wr_data(MESI_RAM_wr_data[3:0]), 
	.rd(MESI_RAM_rd), .rd_addr(MESI_RAM_rd_addr[5:0]), .rd_data(MESI_RAM_rd_data[3:0]) );

wire TAG_RAM_clk;
wire TAG_RAM_res;
reg TAG_RAM_wr;
reg [5:0] TAG_RAM_wr_addr;
reg [23:0] TAG_RAM_wr_data;
reg TAG_RAM_rd;
reg [5:0] TAG_RAM_rd_addr;
reg TAG_RAM_status;
wire [23:0] TAG_RAM_rd_data;
assign TAG_RAM_clk = clk;
assign TAG_RAM_res_n = res_n;
RAM #(24,64) TAG_RAM(.clk(TAG_RAM_clk), .res_n(TAG_RAM_res), 
	.wr(TAG_RAM_wr), .wr_addr(TAG_RAM_wr_addr[5:0]), .wr_data(TAG_RAM_wr_data[23:0]), 
	.rd(TAG_RAM_rd), .rd_addr(TAG_RAM_rd_addr[5:0]), .rd_data(TAG_RAM_rd_data[23:0]) );


  //
 // constant expressions 
//---------------------------
wire IDLE; assign IDLE = ~res_n | (~rd & ~wr);
wire IS_EXCLUSIVE; assign IS_EXCLUSIVE = POP_MESI_RAM(addr[7:2])==EXCLUSIVE;
wire IS_INVALID; assign IS_INVALID = POP_MESI_RAM(addr[7:2])==INVALID;
wire IS_MODIFIED; assign IS_MODIFIED = POP_MESI_RAM(addr[7:2])==MODIFIED;
wire IS_SHARED; assign IS_SHARED = POP_MESI_RAM(addr[7:2])==SHARED;
wire MESI_IS_ZERO; assign MESI_IS_ZERO = POP_MESI_RAM(addr[7:2])==ZERO;
wire READ; assign READ = res_n & cp & rd;
wire TAG_MATCH; assign TAG_MATCH = POP_TAG_RAM(addr[7:2])==addr[31:8];
wire WRITE; assign WRITE = res_n & cp & wr;
wire rREAD; assign rREAD = res_n & ~cp & rd;
wire rWRITE; assign rWRITE = res_n & ~cp & wr;


  //
 // initials for simulator
//---------------------------
initial begin
	STATE <= ID;
	MESI_RAM_wr <= 1'b0;
	MESI_RAM_wr_addr <= {5{1'b0}};
	MESI_RAM_wr_data <= {3{1'b0}};
	MESI_RAM_rd <= 1'b0;
	MESI_RAM_rd_addr <= {5{1'b0}};
	MESI_RAM_status <= 1'b0;
	TAG_RAM_wr <= 1'b0;
	TAG_RAM_wr_addr <= {5{1'b0}};
	TAG_RAM_wr_data <= {23{1'b0}};
	TAG_RAM_rd <= 1'b0;
	TAG_RAM_rd_addr <= {5{1'b0}};
	TAG_RAM_status <= 1'b0;
end


  //
 // state transfer
//---------------------------
always @( posedge clk or negedge res_n) begin
	if( ~res_n ) begin
		STATE <= ID;
	end else begin
		STATE <= NEXT_STATE;
	end // if
end // always@

always @( * ) begin
		case ( STATE )
		ID : begin
			if ( IDLE ) begin 
				NEXT_STATE <= ID;
				// MEMORY action ...
				TAG_RAM_status <= NOP_TAG_RAM(1'b0,ZERO_ADDR);
				MESI_RAM_status <= NOP_MESI_RAM(1'b0,ZERO_ADDR);
				// MEALY output action ...
				mesi<=ZERO;
				index<=addr[7:2];
			end
			if ( READ & TAG_MATCH & ~MESI_IS_ZERO ) begin 
				NEXT_STATE <= RD;
				// MEMORY action ...
				MESI_RAM_status <= PUSH_MESI_RAM(addr[7:2],EXCLUSIVE);
				TAG_RAM_status <= NOP_TAG_RAM(1'b0,ZERO_ADDR);
				// MEALY output action ...
				mesi<=EXCLUSIVE;
				index<=addr[7:2];
			end
			if ( READ & ( ~TAG_MATCH | MESI_IS_ZERO ) ) begin 
				NEXT_STATE <= RD;
				// MEMORY action ...
				MESI_RAM_status <= PUSH_MESI_RAM(addr[7:2],EXCLUSIVE);
				TAG_RAM_status <= PUSH_TAG_RAM(addr[7:2],addr[31:8]);
				// MEALY output action ...
				mesi<=EXCLUSIVE;
				index<=addr[7:2];
			end
			if ( WRITE & TAG_MATCH ) begin 
				NEXT_STATE <= WR;
				// MEMORY action ...
				MESI_RAM_status <= PUSH_MESI_RAM(addr[7:2],MODIFIED);
				TAG_RAM_status <= NOP_TAG_RAM(1'b0,ZERO_ADDR);
				// MEALY output action ...
				mesi<=MODIFIED;
				index<=addr[7:2];
			end
			if ( rWRITE & TAG_MATCH ) begin 
				NEXT_STATE <= rWR;
				// MEMORY action ...
				MESI_RAM_status <= PUSH_MESI_RAM(addr[7:2],INVALID);
				TAG_RAM_status <= NOP_TAG_RAM(1'b0,ZERO_ADDR);
				// MEALY output action ...
				mesi<=INVALID;
				index<=addr[7:2];
			end
			if ( rREAD & TAG_MATCH ) begin 
				NEXT_STATE <= rRD;
				// MEMORY action ...
				MESI_RAM_status <= PUSH_MESI_RAM(addr[7:2],SHARED);
				TAG_RAM_status <= NOP_TAG_RAM(1'b0,ZERO_ADDR);
				// MEALY output action ...
				mesi<=SHARED;
				index<=addr[7:2];
			end
			if ( rREAD & ~TAG_MATCH ) begin 
				NEXT_STATE <= ID;
				// MEMORY action ...
				MESI_RAM_status <= NOP_MESI_RAM(1'b0,ZERO_ADDR);
				TAG_RAM_status <= NOP_TAG_RAM(1'b0,ZERO_ADDR);
				// MEALY output action ...
				mesi<=ZERO;
				index<=addr[7:2];
			end
		end // ID
		RD : begin
			if ( IDLE ) begin 
				NEXT_STATE <= ID;
				// MEMORY action ...
				TAG_RAM_status <= NOP_TAG_RAM(1'b0,ZERO_ADDR);
				MESI_RAM_status <= NOP_MESI_RAM(1'b0,ZERO_ADDR);
				// MEALY output action ...
				mesi<=ZERO;
			end
			if ( READ & TAG_MATCH & ~MESI_IS_ZERO ) begin 
				NEXT_STATE <= RD;
				// MEMORY action ...
				MESI_RAM_status <= PUSH_MESI_RAM(addr[7:2],EXCLUSIVE);
				TAG_RAM_status <= NOP_TAG_RAM(1'b0,ZERO_ADDR);
				// MEALY output action ...
				mesi<=EXCLUSIVE;
				index<=addr[7:2];
			end
			if ( READ & (~TAG_MATCH | MESI_IS_ZERO) ) begin 
				NEXT_STATE <= RD;
				// MEMORY action ...
				MESI_RAM_status <= PUSH_MESI_RAM(addr[7:2],EXCLUSIVE);
				TAG_RAM_status <= PUSH_TAG_RAM(addr[7:2],addr[31:8]);
				// MEALY output action ...
				mesi<=EXCLUSIVE;
				index<=addr[7:2];
			end
		end // RD
		WR : begin
			if ( IDLE ) begin 
				NEXT_STATE <= ID;
				// MEMORY action ...
				TAG_RAM_status <= NOP_TAG_RAM(1'b0,ZERO_ADDR);
				MESI_RAM_status <= NOP_MESI_RAM(1'b0,ZERO_ADDR);
				// MEALY output action ...
				mesi<=ZERO;
				index<=addr[7:2];
			end
			if ( WRITE & TAG_MATCH ) begin 
				NEXT_STATE <= WR;
				// MEMORY action ...
				MESI_RAM_status <= PUSH_MESI_RAM(addr[7:2],MODIFIED);
				TAG_RAM_status <= NOP_TAG_RAM(1'b0,ZERO_ADDR);
				// MEALY output action ...
				mesi<=MODIFIED;
				index<=addr[7:2];
			end
			if ( WRITE & ~TAG_MATCH ) begin 
				NEXT_STATE <= ID;
				// MEMORY action ...
				TAG_RAM_status <= NOP_TAG_RAM(1'b0,ZERO_ADDR);
				MESI_RAM_status <= NOP_MESI_RAM(1'b0,ZERO_ADDR);
				// MEALY output action ...
				mesi<=ZERO;
				index<=addr[7:2];
			end
		end // WR
		rRD : begin
			if ( IDLE ) begin 
				NEXT_STATE <= ID;
				// MEMORY action ...
				TAG_RAM_status <= NOP_TAG_RAM(1'b0,ZERO_ADDR);
				MESI_RAM_status <= NOP_MESI_RAM(1'b0,ZERO_ADDR);
				// MEALY output action ...
				mesi<=ZERO;
				index<=addr[7:2];
			end
			if ( rREAD & TAG_MATCH & ~MESI_IS_ZERO ) begin 
				NEXT_STATE <= rRD;
				// MEMORY action ...
				MESI_RAM_status <= PUSH_MESI_RAM(addr[7:2],SHARED);
				TAG_RAM_status <= NOP_TAG_RAM(1'b0,ZERO_ADDR);
				// MEALY output action ...
				mesi<=SHARED;
				index<=addr[7:2];
			end
			if ( rREAD & ( ~TAG_MATCH | ( TAG_MATCH & MESI_IS_ZERO ) ) ) begin 
				NEXT_STATE <= ID;
				// MEMORY action ...
				TAG_RAM_status <= NOP_TAG_RAM(1'b0,ZERO_ADDR);
				MESI_RAM_status <= NOP_MESI_RAM(1'b0,ZERO_ADDR);
				// MEALY output action ...
				mesi<=ZERO;
				index<=addr[7:2];
			end
		end // rRD
		rWR : begin
			if ( IDLE ) begin 
				NEXT_STATE <= ID;
				// MEMORY action ...
				TAG_RAM_status <= NOP_TAG_RAM(1'b0,ZERO_ADDR);
				MESI_RAM_status <= NOP_MESI_RAM(1'b0,ZERO_ADDR);
				// MEALY output action ...
				mesi<=ZERO;
				index<=addr[7:2];
			end
			if ( rWRITE & TAG_MATCH & ~MESI_IS_ZERO ) begin 
				NEXT_STATE <= rWR;
				// MEMORY action ...
				MESI_RAM_status <= PUSH_MESI_RAM(addr[7:2],INVALID);
				TAG_RAM_status <= NOP_TAG_RAM(1'b0,ZERO_ADDR);
				// MEALY output action ...
				mesi<=INVALID;
				index<=addr[7:2];
			end
			if ( rWRITE & ~TAG_MATCH | (TAG_MATCH & MESI_IS_ZERO) ) begin 
				NEXT_STATE <= rWR;
				// MEMORY action ...
				MESI_RAM_status <= NOP_MESI_RAM(1'b0,ZERO_ADDR);
				TAG_RAM_status <= NOP_TAG_RAM(1'b0,ZERO_ADDR);
				// MEALY output action ...
				mesi<=ZERO;
				index<=addr[7:2];
			end
		end // rWR
		default begin 
			NEXT_STATE <= ID;
			mesi[3:0] <= {4{1'b0}};
			index[5:0] <= {6{1'b0}};
		end
		endcase
end // always@


  //
 // memory functions
//---------------------------
function [3:0] POP_MESI_RAM;
   input [5:0] rd_addr;
begin
   MESI_RAM_rd = 1'b1;
   MESI_RAM_rd_addr = rd_addr;
   POP_MESI_RAM = MESI_RAM_rd_data;
end
endfunction

function PUSH_MESI_RAM;
   input [5:0] wr_addr;
   input [3:0] wr_data;
begin
   MESI_RAM_wr = 1'b1;
   MESI_RAM_wr_addr = wr_addr;
   MESI_RAM_wr_data = wr_data;
   PUSH_MESI_RAM = 1'b1;
end
endfunction

function TOP_MESI_RAM;
   input [5:0] rd_addr;
begin
   MESI_RAM_rd = 1'b1;
   MESI_RAM_rd_addr = rd_addr;
   TOP_MESI_RAM = MESI_RAM_rd_data;
end
endfunction

function NOP_MESI_RAM;
   input wr;
   input [5:0] wr_addr;
begin
   //MESI_RAM_rd = 1'b0;
   MESI_RAM_wr = wr;
   MESI_RAM_wr_addr = wr_addr;
   MESI_RAM_wr_data = {4{1'b0}};
   NOP_MESI_RAM = 1'b1;
end
endfunction

function [23:0] POP_TAG_RAM;
   input [5:0] rd_addr;
begin
   TAG_RAM_rd = 1'b1;
   TAG_RAM_rd_addr = rd_addr;
   POP_TAG_RAM = TAG_RAM_rd_data;
end
endfunction

function PUSH_TAG_RAM;
   input [5:0] wr_addr;
   input [23:0] wr_data;
begin
   TAG_RAM_wr = 1'b1;
   TAG_RAM_wr_addr = wr_addr;
   TAG_RAM_wr_data = wr_data;
   PUSH_TAG_RAM = 1'b1;
end
endfunction

function TOP_TAG_RAM;
   input [5:0] rd_addr;
begin
   TAG_RAM_rd = 1'b1;
   TAG_RAM_rd_addr = rd_addr;
   TOP_TAG_RAM = TAG_RAM_rd_data;
end
endfunction

function NOP_TAG_RAM;
   input wr;
   input [5:0] wr_addr;
begin
   //TAG_RAM_rd = 1'b0;
   TAG_RAM_wr = wr;
   TAG_RAM_wr_addr = wr_addr;
   TAG_RAM_wr_data = {24{1'b0}};
   NOP_TAG_RAM = 1'b1;
end
endfunction

endmodule


  //
 // memory
//---------------------------
module RAM ( clk, res_n, wr, wr_addr, wr_data, rd, rd_addr, rd_data );

parameter data_width = 32;
parameter entries = 16;
parameter addr_width = $clog2(entries);

input  wire clk, res_n, rd, wr;
input  wire [data_width-1:0] wr_data;
input  wire [addr_width-1:0] rd_addr, wr_addr;
output wire [data_width-1:0] rd_data;

parameter ZERO_ADDR = {addr_width{1'b0}};
parameter ZERO_DATA = {data_width{1'b0}};
parameter ONE_DATA = {data_width{1'b1}};
    
reg    [data_width-1:0] MEMORY [entries-1:0];

integer i;

initial begin
   // Clear RAM before simulation start
   for ( i = 0; i < entries; i = i + 1 ) begin
       MEMORY[ i ] <= ZERO_DATA;
   end
end

// IN-CYCLE READ ACCESS
assign rd_data = {data_width{rd & res_n}} & MEMORY[rd_addr];

always @ ( posedge clk ) begin
    // WRITE ACCESS
   MEMORY[wr_addr] <= {data_width{wr & res_n}} & wr_data;
end

endmodule
